@@ -1,6 +1,6 @@
 // SPDX-License-Identifier: GPL-2.0 OR Linux-OpenIB
 /*
- * Copyright 2018-2024 Amazon.com, Inc. or its affiliates. All rights reserved.
+ * Copyright 2018-2025 Amazon.com, Inc. or its affiliates. All rights reserved.
  */
 
 #include "kcompat.h"
@@ -23,6 +23,9 @@
 #endif
 
 #include "efa.h"
+#if defined(HAVE_EFA_KVERBS) && !defined(IN_TREE_DRIVER)
+#include "efa_verbs.h"
+#endif
 #include "efa_io_defs.h"
 
 #ifdef HAVE_EFA_P2P
@@ -35,10 +38,6 @@ enum {
 	EFA_MMAP_IO_NC,
 };
 
-#define EFA_AENQ_ENABLED_GROUPS \
-	(BIT(EFA_ADMIN_FATAL_ERROR) | BIT(EFA_ADMIN_WARNING) | \
-	 BIT(EFA_ADMIN_NOTIFICATION) | BIT(EFA_ADMIN_KEEP_ALIVE))
-
 struct efa_user_mmap_entry {
 	struct rdma_user_mmap_entry rdma_entry;
 #ifndef HAVE_CORE_MMAP_XA
@@ -62,6 +61,12 @@ struct efa_user_mmap_entry {
 	op(EFA_CREATE_AH_ERR, "create_ah_err") \
 	op(EFA_MMAP_ERR, "mmap_err")
 
+#ifdef HAVE_EFA_KVERBS
+#define EFA_DEFINE_KVERBS_DEVICE_STATS(op) \
+	op(EFA_ALLOC_MR_ERR, "alloc_mr_err") \
+	op(EFA_GET_DMA_MR_ERR, "get_dma_mr_err")
+#endif
+
 #define EFA_DEFINE_PORT_STATS(op) \
 	op(EFA_TX_BYTES, "tx_bytes") \
 	op(EFA_TX_PKTS, "tx_pkts") \
@@ -92,6 +97,9 @@ struct efa_user_mmap_entry {
 
 enum efa_hw_device_stats {
 	EFA_DEFINE_DEVICE_STATS(EFA_STATS_ENUM)
+#ifdef HAVE_EFA_KVERBS
+	EFA_DEFINE_KVERBS_DEVICE_STATS(EFA_STATS_ENUM)
+#endif
 };
 
 #ifdef HAVE_STAT_DESC_STRUCT
@@ -100,6 +108,9 @@ static const struct rdma_stat_desc efa_device_stats_descs[] = {
 static const char *const efa_device_stats_descs[] = {
 #endif
 	EFA_DEFINE_DEVICE_STATS(EFA_STATS_STR)
+#ifdef HAVE_EFA_KVERBS
+	EFA_DEFINE_KVERBS_DEVICE_STATS(EFA_STATS_STR)
+#endif
 };
 
 enum efa_hw_port_stats {
@@ -114,6 +125,8 @@ static const char *const efa_port_stats_descs[] = {
 	EFA_DEFINE_PORT_STATS(EFA_STATS_STR)
 };
 
+#define EFA_DEFAULT_LINK_SPEED_GBPS   100
+
 #define EFA_CHUNK_PAYLOAD_SHIFT       12
 #define EFA_CHUNK_PAYLOAD_SIZE        BIT(EFA_CHUNK_PAYLOAD_SHIFT)
 #define EFA_CHUNK_PAYLOAD_PTR_SIZE    8
@@ -382,10 +395,47 @@ int efa_query_device(struct ib_device *ibdev,
 	return 0;
 }
 
+static void efa_link_gbps_to_speed_and_width(u16 gbps,
+					     enum ib_port_speed *speed,
+					     enum ib_port_width *width)
+{
+	if (gbps >= 400) {
+		*width = IB_WIDTH_8X;
+		*speed = IB_SPEED_HDR;
+	} else if (gbps >= 200) {
+		*width = IB_WIDTH_4X;
+		*speed = IB_SPEED_HDR;
+	} else if (gbps >= 120) {
+		*width = IB_WIDTH_12X;
+		*speed = IB_SPEED_FDR10;
+	} else if (gbps >= 100) {
+		*width = IB_WIDTH_4X;
+		*speed = IB_SPEED_EDR;
+	} else if (gbps >= 60) {
+		*width = IB_WIDTH_12X;
+		*speed = IB_SPEED_DDR;
+	} else if (gbps >= 50) {
+		*width = IB_WIDTH_1X;
+		*speed = IB_SPEED_HDR;
+	} else if (gbps >= 40) {
+		*width = IB_WIDTH_4X;
+		*speed = IB_SPEED_FDR10;
+	} else if (gbps >= 30) {
+		*width = IB_WIDTH_12X;
+		*speed = IB_SPEED_SDR;
+	} else {
+		*width = IB_WIDTH_1X;
+		*speed = IB_SPEED_EDR;
+	}
+}
+
 int efa_query_port(struct ib_device *ibdev, port_t port,
 		   struct ib_port_attr *props)
 {
 	struct efa_dev *dev = to_edev(ibdev);
+	enum ib_port_speed link_speed;
+	enum ib_port_width link_width;
+	u16 link_gbps;
 
 	props->lmc = 1;
 
@@ -393,8 +443,10 @@ int efa_query_port(struct ib_device *ibdev, port_t port,
 	props->phys_state = IB_PORT_PHYS_STATE_LINK_UP;
 	props->gid_tbl_len = 1;
 	props->pkey_tbl_len = 1;
-	props->active_speed = IB_SPEED_EDR;
-	props->active_width = IB_WIDTH_4X;
+	link_gbps = dev->dev_attr.max_link_speed_gbps ?: EFA_DEFAULT_LINK_SPEED_GBPS;
+	efa_link_gbps_to_speed_and_width(link_gbps, &link_speed, &link_width);
+	props->active_speed = link_speed;
+	props->active_width = link_width;
 	props->max_mtu = ib_mtu_int_to_enum(dev->dev_attr.mtu);
 	props->active_mtu = ib_mtu_int_to_enum(dev->dev_attr.mtu);
 	props->max_msg_sz = dev->dev_attr.mtu;
@@ -497,15 +549,11 @@ int efa_alloc_pd(struct ib_pd *ibpd,
 	struct efa_pd *pd = to_epd(ibpd);
 	int err;
 
-#ifndef HAVE_NO_KVERBS_DRIVERS
-	if (!udata) {
-		ibdev_dbg(&dev->ibdev, "udata is NULL\n");
-		err = -EOPNOTSUPP;
-		goto err_out;
-	}
-#endif
-
+#ifdef HAVE_EFA_KVERBS
+	if (udata && udata->inlen &&
+#else
 	if (udata->inlen &&
+#endif
 	    !ib_is_udata_cleared(udata, 0, udata->inlen)) {
 		ibdev_dbg(&dev->ibdev,
 			  "Incompatible ABI params, udata not cleared\n");
@@ -520,7 +568,11 @@ int efa_alloc_pd(struct ib_pd *ibpd,
 	pd->pdn = result.pdn;
 	resp.pdn = result.pdn;
 
+#ifdef HAVE_EFA_KVERBS
+	if (udata && udata->outlen) {
+#else
 	if (udata->outlen) {
+#endif
 		err = ib_copy_to_udata(udata, &resp,
 				       min(sizeof(resp), udata->outlen));
 		if (err) {
@@ -598,6 +650,65 @@ int efa_dealloc_pd(struct ib_pd *ibpd)
 #endif
 }
 
+#ifdef HAVE_EFA_KVERBS
+static void efa_lock_cqs(struct efa_cq *send_cq, struct efa_cq *recv_cq)
+{
+	if (recv_cq == send_cq) {
+		spin_lock(&recv_cq->lock);
+	} else {
+		spin_lock(&recv_cq->lock);
+		spin_lock(&send_cq->lock);
+	}
+}
+
+static void efa_unlock_cqs(struct efa_cq *send_cq, struct efa_cq *recv_cq)
+{
+	if (recv_cq == send_cq) {
+		spin_unlock(&recv_cq->lock);
+	} else {
+		spin_unlock(&recv_cq->lock);
+		spin_unlock(&send_cq->lock);
+	}
+}
+
+static void efa_cq_inc_ref_cnt(struct efa_cq *cq, u8 sub_cq_idx)
+{
+	cq->sub_cq_arr[sub_cq_idx].ref_cnt++;
+}
+
+static void efa_cq_dec_ref_cnt(struct efa_cq *cq, u8 sub_cq_idx)
+{
+	cq->sub_cq_arr[sub_cq_idx].ref_cnt--;
+}
+
+static void efa_wq_terminate(struct efa_wq *wq)
+{
+	kfree(wq->wrid_idx_pool);
+	kfree(wq->wrid);
+}
+
+static void efa_qp_terminate(struct efa_qp *qp)
+{
+	struct efa_dev *dev = to_edev(qp->ibqp.pd->device);
+	struct efa_cq *send_cq = to_ecq(qp->ibqp.send_cq);
+	struct efa_cq *recv_cq = to_ecq(qp->ibqp.recv_cq);
+
+	spin_lock(&dev->qp_table_lock);
+	efa_lock_cqs(send_cq, recv_cq);
+
+	efa_cq_dec_ref_cnt(send_cq, qp->sq.wq.sub_cq_idx);
+	efa_cq_dec_ref_cnt(recv_cq, qp->rq.wq.sub_cq_idx);
+
+	dev->qp_table[qp->ibqp.qp_num] = NULL;
+
+	efa_unlock_cqs(send_cq, recv_cq);
+	spin_unlock(&dev->qp_table_lock);
+
+	efa_wq_terminate(&qp->sq.wq);
+	efa_wq_terminate(&qp->rq.wq);
+}
+#endif
+
 static int efa_destroy_qp_handle(struct efa_dev *dev, u32 qp_handle)
 {
 	struct efa_com_destroy_qp_params params = { .qp_handle = qp_handle };
@@ -629,7 +740,14 @@ int efa_destroy_qp(struct ib_qp *ibqp)
 	if (err)
 		return err;
 
+#ifdef HAVE_EFA_KVERBS
+	if (ibqp->uobject)
+		efa_qp_user_mmap_entries_remove(qp);
+	else
+		efa_qp_terminate(qp);
+#else
 	efa_qp_user_mmap_entries_remove(qp);
+#endif
 
 	if (qp->rq_cpu_addr) {
 		ibdev_dbg(&dev->ibdev,
@@ -735,7 +853,7 @@ static int qp_mmap_entries_setup(struct efa_qp *qp,
 
 	address = dev->mem_bar_addr + resp->llq_desc_offset;
 	length = PAGE_ALIGN(params->sq_ring_size_in_bytes +
-			    (resp->llq_desc_offset & ~PAGE_MASK));
+			    offset_in_page(resp->llq_desc_offset));
 
 	qp->llq_desc_mmap_entry =
 		efa_user_mmap_entry_insert(&ucontext->ibucontext,
@@ -843,6 +961,250 @@ static int efa_qp_validate_attr(struct efa_dev *dev,
 	return 0;
 }
 
+#ifdef HAVE_EFA_KVERBS
+static int efa_qp_validate_attr_kernel(struct efa_dev *dev, struct ib_qp_init_attr *init_attr)
+{
+	if (init_attr->qp_type != EFA_QPT_SRD && init_attr->qp_type != IB_QPT_UD) {
+		ibdev_dbg(&dev->ibdev, "Unsupported qp type %d\n", init_attr->qp_type);
+		return -EOPNOTSUPP;
+	}
+
+	if (!init_attr->send_cq || !init_attr->recv_cq) {
+		ibdev_dbg(&dev->ibdev, "Send/Receive CQ wasn't provided\n");
+		return -EINVAL;
+	}
+
+	if (init_attr->srq) {
+		ibdev_dbg(&dev->ibdev, "SRQ is not supported\n");
+		return -EOPNOTSUPP;
+	}
+
+	if (init_attr->create_flags) {
+		ibdev_dbg(&dev->ibdev, "Unsupported create flags\n");
+		return -EOPNOTSUPP;
+	}
+
+	return 0;
+}
+
+static void efa_qp_init_indices(struct efa_qp *qp)
+{
+	qp->sq.wq.wqes_posted = 0;
+	qp->sq.wq.wqes_completed = 0;
+	qp->sq.wq.pc = 0;
+	qp->sq.wq.wrid_idx_pool_next = 0;
+
+	qp->rq.wq.wqes_posted = 0;
+	qp->rq.wq.wqes_completed = 0;
+	qp->rq.wq.pc = 0;
+	qp->rq.wq.wrid_idx_pool_next = 0;
+}
+
+static void efa_setup_qp(struct efa_qp *qp, struct ib_qp_cap *cap)
+{
+	u32 rq_desc_cnt;
+
+	efa_qp_init_indices(qp);
+
+	qp->sq.wq.max_wqes = roundup_pow_of_two(cap->max_send_wr);
+	qp->sq.wq.max_sge = cap->max_send_sge;
+	qp->sq.wq.queue_mask = qp->sq.wq.max_wqes - 1;
+
+	qp->rq.wq.max_sge = cap->max_recv_sge;
+	rq_desc_cnt = roundup_pow_of_two(cap->max_recv_sge * cap->max_recv_wr);
+	qp->rq.wq.queue_mask = rq_desc_cnt - 1;
+	qp->rq.wq.max_wqes = rq_desc_cnt / qp->rq.wq.max_sge;
+}
+
+static int efa_wq_initialize(struct efa_dev *dev, struct efa_wq *wq, u32 db_off, u16 sub_cq_idx)
+{
+	int err;
+	int i;
+
+	wq->wrid = kvzalloc(wq->max_wqes * sizeof(*wq->wrid), GFP_KERNEL);
+	if (!wq->wrid)
+		return -ENOMEM;
+
+	wq->wrid_idx_pool = kvzalloc(wq->max_wqes * sizeof(*wq->wrid_idx_pool), GFP_KERNEL);
+	if (!wq->wrid_idx_pool) {
+		err = -ENOMEM;
+		goto err_free_wrid;
+	}
+
+	wq->db = (u32 *)(dev->db_bar + db_off);
+
+	/* Initialize the wrid free indexes pool. */
+	for (i = 0; i < wq->max_wqes; i++)
+		wq->wrid_idx_pool[i] = i;
+
+	spin_lock_init(&wq->lock);
+
+	wq->sub_cq_idx = sub_cq_idx;
+
+	return 0;
+
+err_free_wrid:
+	kvfree(wq->wrid);
+	return err;
+}
+
+static int efa_rq_initialize(struct efa_dev *dev,
+			     struct efa_qp *qp,
+			     struct efa_com_create_qp_result *res)
+{
+	struct efa_rq *rq = &qp->rq;
+	int err;
+
+	if (!rq->wq.max_wqes)
+		return 0;
+
+	err = efa_wq_initialize(dev, &qp->rq.wq, res->rq_db_offset, res->recv_sub_cq_idx);
+	if (err)
+		return err;
+
+	rq->buf_size = qp->rq_size;
+	rq->buf = (u8 *)qp->rq_cpu_addr;
+
+	return 0;
+}
+
+static int efa_sq_initialize(struct efa_dev *dev,
+			     struct efa_qp *qp,
+			     const struct ib_qp_init_attr *attr,
+			     struct efa_com_create_qp_result *res)
+{
+	struct efa_sq *sq = &qp->sq;
+	int err;
+
+	if (!sq->wq.max_wqes)
+		return 0;
+
+	err = efa_wq_initialize(dev, &qp->sq.wq, res->sq_db_offset, res->send_sub_cq_idx);
+	if (err)
+		return err;
+
+	sq->desc_offset = res->llq_descriptors_offset;
+	sq->desc = (u8 *)(dev->mem_bar + res->llq_descriptors_offset);
+
+	sq->max_inline_data = attr->cap.max_inline_data;
+	sq->max_rdma_sges = 1;
+	sq->max_batch_wr = dev->dev_attr.max_tx_batch ?
+		(dev->dev_attr.max_tx_batch * 64) / sizeof(struct efa_io_tx_wqe) : U16_MAX;
+
+	if (dev->dev_attr.min_sq_depth) {
+		/* The device can't accept a doorbell for the whole SQ at once,
+		 * set the max batch to at least (SQ size - 1).
+		 */
+		sq->max_batch_wr = min_t(u32, sq->max_batch_wr, sq->wq.max_wqes - 1);
+	}
+
+	sq->sig_type = attr->sq_sig_type;
+
+	return 0;
+}
+
+static int efa_create_qp_kernel(struct ib_qp *ibqp, struct ib_qp_init_attr *init_attr)
+{
+	struct efa_com_create_qp_params create_qp_params = {};
+	struct efa_cq *send_cq = to_ecq(init_attr->send_cq);
+	struct efa_cq *recv_cq = to_ecq(init_attr->recv_cq);
+	struct efa_com_create_qp_result create_qp_resp;
+	struct efa_dev *dev = to_edev(ibqp->device);
+	struct efa_qp *qp = to_eqp(ibqp);
+	int err;
+
+	err = efa_qp_validate_cap(dev, init_attr);
+	if (err)
+		goto err_out;
+
+	err = efa_qp_validate_attr_kernel(dev, init_attr);
+	if (err)
+		goto err_out;
+
+	efa_setup_qp(qp, &init_attr->cap);
+
+	create_qp_params.uarn = dev->uarn;
+	create_qp_params.pd = to_epd(ibqp->pd)->pdn;
+
+	if (init_attr->qp_type == IB_QPT_UD) {
+		create_qp_params.qp_type = EFA_ADMIN_QP_TYPE_UD;
+	} else if (init_attr->qp_type == EFA_QPT_SRD) {
+		create_qp_params.qp_type = EFA_ADMIN_QP_TYPE_SRD;
+	} else {
+		ibdev_dbg(&dev->ibdev, "Unsupported qp type %d\n", init_attr->qp_type);
+		err = -EOPNOTSUPP;
+		goto err_out;
+	}
+
+	create_qp_params.send_cq_idx = send_cq->cq_idx;
+	create_qp_params.recv_cq_idx = recv_cq->cq_idx;
+	create_qp_params.sq_depth = init_attr->cap.max_send_wr;
+	create_qp_params.sq_ring_size_in_bytes = (init_attr->cap.max_send_wr) *
+			sizeof(struct efa_io_tx_wqe);
+
+	create_qp_params.rq_depth = init_attr->cap.max_recv_wr;
+	create_qp_params.rq_ring_size_in_bytes = (qp->rq.wq.queue_mask + 1) *
+			sizeof(struct efa_io_rx_desc);
+	qp->rq_size = PAGE_ALIGN(create_qp_params.rq_ring_size_in_bytes);
+	if (qp->rq_size) {
+		qp->rq_cpu_addr = efa_zalloc_mapped(dev, &qp->rq_dma_addr,
+						    qp->rq_size, DMA_TO_DEVICE);
+		if (!qp->rq_cpu_addr) {
+			err = -ENOMEM;
+			goto err_out;
+		}
+
+		ibdev_dbg(&dev->ibdev, "qp->cpu_addr[0x%p] allocated: size[%lu], dma[%pad]\n",
+			  qp->rq_cpu_addr, qp->rq_size, &qp->rq_dma_addr);
+		create_qp_params.rq_base_addr = qp->rq_dma_addr;
+	}
+
+	err = efa_com_create_qp(&dev->edev, &create_qp_params, &create_qp_resp);
+	if (err)
+		goto err_free_mapped;
+
+	ibqp->qp_num = create_qp_resp.qp_num;
+	qp->qp_handle = create_qp_resp.qp_handle;
+	qp->max_send_wr = init_attr->cap.max_send_wr;
+	qp->max_recv_wr = init_attr->cap.max_recv_wr;
+	qp->max_send_sge = init_attr->cap.max_send_sge;
+	qp->max_recv_sge = init_attr->cap.max_recv_sge;
+	qp->max_inline_data = init_attr->cap.max_inline_data;
+
+	err = efa_rq_initialize(dev, qp, &create_qp_resp);
+	if (err)
+		goto err_destroy_qp;
+
+	err = efa_sq_initialize(dev, qp, init_attr, &create_qp_resp);
+	if (err)
+		goto err_terminate_rq;
+
+	efa_lock_cqs(send_cq, recv_cq);
+	efa_cq_inc_ref_cnt(send_cq, create_qp_resp.send_sub_cq_idx);
+	efa_cq_inc_ref_cnt(recv_cq, create_qp_resp.recv_sub_cq_idx);
+	efa_unlock_cqs(send_cq, recv_cq);
+
+	spin_lock(&dev->qp_table_lock);
+	dev->qp_table[ibqp->qp_num] = qp;
+	spin_unlock(&dev->qp_table_lock);
+
+	ibdev_dbg(&dev->ibdev, "Created qp[%d] type %d\n", qp->ibqp.qp_num, init_attr->qp_type);
+
+	return 0;
+
+err_terminate_rq:
+	efa_wq_terminate(&qp->rq.wq);
+err_destroy_qp:
+	efa_destroy_qp_handle(dev, create_qp_resp.qp_handle);
+err_free_mapped:
+	if (qp->rq_size)
+		efa_free_mapped(dev, qp->rq_cpu_addr, qp->rq_dma_addr, qp->rq_size, DMA_TO_DEVICE);
+err_out:
+	atomic64_inc(&dev->stats.create_qp_err);
+	return err;
+}
+#endif
+
 int efa_create_qp(struct ib_qp *ibqp, struct ib_qp_init_attr *init_attr,
 		  struct ib_udata *udata)
 {
@@ -856,14 +1218,10 @@ int efa_create_qp(struct ib_qp *ibqp, struct ib_qp_init_attr *init_attr,
 	u16 supported_efa_flags = 0;
 	int err;
 
-#ifndef HAVE_NO_KVERBS_DRIVERS
-	if (!udata) {
-		ibdev_dbg(&dev->ibdev, "udata is NULL\n");
-		err = -EOPNOTSUPP;
-		goto err_out;
-	}
+#ifdef HAVE_EFA_KVERBS
+	if (!udata)
+		return efa_create_qp_kernel(ibqp, init_attr);
 #endif
-
 #ifdef HAVE_UDATA_TO_DRV_CONTEXT
 	ucontext = rdma_udata_to_drv_context(udata, struct efa_ucontext,
 					     ibucontext);
@@ -904,7 +1262,7 @@ int efa_create_qp(struct ib_qp *ibqp, struct ib_qp_init_attr *init_attr,
 		goto err_out;
 	}
 
-	if (cmd.comp_mask || !is_reserved_cleared(cmd.reserved_90)) {
+	if (cmd.comp_mask || !is_reserved_cleared(cmd.reserved_98)) {
 		ibdev_dbg(&dev->ibdev,
 			  "Incompatible ABI params, unknown fields in udata\n");
 		err = -EINVAL;
@@ -960,6 +1318,8 @@ int efa_create_qp(struct ib_qp *ibqp, struct ib_qp_init_attr *init_attr,
 		create_qp_params.rq_base_addr = qp->rq_dma_addr;
 	}
 
+	create_qp_params.sl = cmd.sl;
+
 	if (cmd.flags & EFA_CREATE_QP_WITH_UNSOLICITED_WRITE_RECV)
 		create_qp_params.unsolicited_write_recv = true;
 
@@ -1175,7 +1535,11 @@ static int efa_modify_qp_validate(struct efa_dev *dev, struct efa_qp *qp,
 		return -EOPNOTSUPP;
 	}
 
+#ifdef HAVE_EFA_KVERBS
+	if (qp->ibqp.qp_type == IB_QPT_DRIVER || qp->ibqp.qp_type == EFA_QPT_SRD)
+#else
 	if (qp->ibqp.qp_type == IB_QPT_DRIVER)
+#endif
 		err = !efa_modify_srd_qp_is_ok(cur_state, new_state,
 					       qp_attr_mask);
 	else
@@ -1216,19 +1580,16 @@ int efa_modify_qp(struct ib_qp *ibqp, struct ib_qp_attr *qp_attr,
 	enum ib_qp_state new_state;
 	int err;
 
-#ifndef HAVE_NO_KVERBS_DRIVERS
-	if (!udata) {
-		ibdev_dbg(&dev->ibdev, "udata is NULL\n");
-		return -EOPNOTSUPP;
-	}
-#endif
-
 #ifdef HAVE_UVERBS_CMD_MASK_NOT_NEEDED
 	if (qp_attr_mask & ~IB_QP_ATTR_STANDARD_BITS)
 		return -EOPNOTSUPP;
 #endif
 
+#ifdef HAVE_EFA_KVERBS
+	if (udata && udata->inlen &&
+#else
 	if (udata->inlen &&
+#endif
 	    !ib_is_udata_cleared(udata, 0, udata->inlen)) {
 		ibdev_dbg(&dev->ibdev,
 			  "Incompatible ABI params, udata not cleared\n");
@@ -1299,11 +1660,12 @@ static void efa_cq_user_mmap_entries_remove(struct efa_cq *cq)
 	rdma_user_mmap_entry_remove(cq->mmap_entry);
 }
 
-#if defined(HAVE_IB_VOID_DESTROY_CQ) || defined(HAVE_IB_INT_DESTROY_CQ)
-#ifdef HAVE_IB_INT_DESTROY_CQ
+#ifdef HAVE_IB_INT_DESTROY_CQ_UDATA
 int efa_destroy_cq(struct ib_cq *ibcq, struct ib_udata *udata)
-#else
+#elif defined(HAVE_IB_VOID_DESTROY_CQ_UDATA)
 void efa_destroy_cq(struct ib_cq *ibcq, struct ib_udata *udata)
+#else
+int efa_destroy_cq(struct ib_cq *ibcq)
 #endif
 {
 	struct efa_dev *dev = to_edev(ibcq->device);
@@ -1325,49 +1687,16 @@ void efa_destroy_cq(struct ib_cq *ibcq, struct ib_udata *udata)
 	}
 	efa_free_mapped(dev, cq->cpu_addr, cq->dma_addr, cq->size,
 			DMA_FROM_DEVICE);
+#ifdef HAVE_EFA_KVERBS
+	kfree(cq->sub_cq_arr);
+#endif
 #ifndef HAVE_CQ_CORE_ALLOCATION
 	kfree(cq);
 #endif
-#ifdef HAVE_IB_INT_DESTROY_CQ
+#ifndef HAVE_IB_VOID_DESTROY_CQ_UDATA
 	return 0;
 #endif
 }
-#else
-#ifdef HAVE_DESTROY_CQ_UDATA
-int efa_destroy_cq(struct ib_cq *ibcq, struct ib_udata *udata)
-#else
-int efa_destroy_cq(struct ib_cq *ibcq)
-#endif
-{
-	struct efa_dev *dev = to_edev(ibcq->device);
-	struct efa_cq *cq = to_ecq(ibcq);
-	int err;
-
-	ibdev_dbg(&dev->ibdev,
-		  "Destroy cq[%d] virt[0x%p] freed: size[%lu], dma[%pad]\n",
-		  cq->cq_idx, cq->cpu_addr, cq->size, &cq->dma_addr);
-
-	err = efa_destroy_cq_idx(dev, cq->cq_idx);
-	if (err)
-		return err;
-
-	efa_cq_user_mmap_entries_remove(cq);
-
-	if (cq->eq) {
-#ifdef HAVE_XARRAY
-		xa_erase(&dev->cqs_xa, cq->cq_idx);
-#else
-		dev->cqs_arr[cq->cq_idx] = NULL;
-#endif
-		synchronize_irq(cq->eq->irq.irqn);
-	}
-	efa_free_mapped(dev, cq->cpu_addr, cq->dma_addr, cq->size,
-			DMA_FROM_DEVICE);
-
-	kfree(cq);
-	return 0;
-}
-#endif
 
 static struct efa_eq *efa_vec2eq(struct efa_dev *dev, int vec)
 {
@@ -1404,9 +1733,144 @@ static int cq_mmap_entries_setup(struct efa_dev *dev, struct efa_cq *cq,
 	return 0;
 }
 
+#ifdef HAVE_EFA_KVERBS
+static void efa_sub_cq_initialize(struct efa_sub_cq *sub_cq, u8 *buf,
+				  int sub_cq_size, int cqe_size)
+{
+	sub_cq->consumed_cnt = 0;
+	sub_cq->phase = 1;
+	sub_cq->buf = buf;
+	sub_cq->queue_mask = sub_cq_size - 1;
+	sub_cq->cqe_size = cqe_size;
+	sub_cq->ref_cnt = 0;
+}
+
+static int efa_create_cq_kernel(struct ib_cq *ibcq, const struct ib_cq_init_attr *attr)
+{
+	int entries = roundup_pow_of_two(attr->cqe);
+	struct efa_com_create_cq_params params = {};
+	struct efa_com_create_cq_result result;
+	struct ib_device *ibdev = ibcq->device;
+	struct efa_dev *dev = to_edev(ibdev);
+	struct efa_cq *cq = to_ecq(ibcq);
+	bool set_src_addr = false;
+	int sub_buf_size;
+	int num_sub_cqs;
+	u32 entry_size;
+	int err = 0;
+	u8 *buf;
+	int i;
+
+	ibdev_dbg(ibdev, "Create_cq entries %d\n", entries);
+
+	if (attr->flags)
+		return -EOPNOTSUPP;
+
+	if (entries < 1 || entries > dev->dev_attr.max_cq_depth) {
+		ibdev_dbg(ibdev,
+			  "cq: requested entries[%u] non-positive or greater than max[%u]\n",
+			  entries, dev->dev_attr.max_cq_depth);
+		err = -EINVAL;
+		goto err_out;
+	}
+
+	num_sub_cqs = dev->dev_attr.sub_cqs_per_cq;
+	cq->sub_cq_arr = kvzalloc(num_sub_cqs * sizeof(*cq->sub_cq_arr), GFP_KERNEL);
+	if (!cq->sub_cq_arr) {
+		err = -ENOMEM;
+		goto err_out;
+	}
+
+	entry_size = sizeof(struct efa_io_rx_cdesc);
+	cq->size = PAGE_ALIGN(entry_size * entries * num_sub_cqs);
+	cq->cpu_addr = efa_zalloc_mapped(dev, &cq->dma_addr, cq->size, DMA_FROM_DEVICE);
+	if (!cq->cpu_addr) {
+		err = -ENOMEM;
+		goto err_free_sub_cq_arr;
+	}
+
+	params.uarn = dev->uarn;
+	params.sub_cq_depth = entries;
+	params.dma_addr = cq->dma_addr;
+	params.entry_size_in_bytes = entry_size;
+	params.num_sub_cqs = num_sub_cqs;
+	params.set_src_addr = set_src_addr;
+
+	if (ibcq->comp_handler) {
+		if (attr->comp_vector >= dev->neqs) {
+			ibdev_dbg(ibdev, "Invalid comp vector request[%u]\n", attr->comp_vector);
+			err = -EINVAL;
+			goto err_free_mapped;
+		}
+
+		cq->eq = efa_vec2eq(dev, attr->comp_vector);
+	}
+
+	if (cq->eq) {
+		params.eqn = cq->eq->eeq.eqn;
+		params.interrupt_mode_enabled = true;
+		ibdev_dbg(ibdev, "Create cq with events eqn: %d\n", params.eqn);
+	}
+
+	err = efa_com_create_cq(&dev->edev, &params, &result);
+	if (err)
+		goto err_free_mapped;
+
+	cq->cq_idx = result.cq_idx;
+	cq->ibcq.cqe = result.actual_depth;
+	WARN_ON_ONCE(entries != result.actual_depth);
+
+	cq->buf_size = cq->size;
+	cq->num_sub_cqs = num_sub_cqs;
+	cq->db = (u32 *)(dev->db_bar + result.db_off);
+
+	cq->buf = cq->cpu_addr;
+	buf = cq->buf;
+	sub_buf_size = entry_size * result.actual_depth;
+	for (i = 0; i < cq->num_sub_cqs; i++) {
+		efa_sub_cq_initialize(&cq->sub_cq_arr[i], buf, result.actual_depth, entry_size);
+		buf += sub_buf_size;
+	}
+
+	if (cq->eq) {
+#ifdef HAVE_XARRAY
+		err = xa_err(xa_store(&dev->cqs_xa, cq->cq_idx, cq, GFP_KERNEL));
+#else
+		dev->cqs_arr[cq->cq_idx] = cq;
+#endif
+		if (err) {
+			ibdev_dbg(ibdev, "Failed to store cq[%u] in xarray\n", cq->cq_idx);
+			goto err_destroy_cq;
+		}
+	}
+
+	spin_lock_init(&cq->lock);
+
+	ibdev_dbg(ibdev, "Created cq[%d], cq depth[%u]. dma[%pad] virt[0x%p]\n",
+		  cq->cq_idx, result.actual_depth, &cq->dma_addr, cq->cpu_addr);
+
+	return 0;
+
+err_destroy_cq:
+	efa_destroy_cq_idx(dev, cq->cq_idx);
+err_free_mapped:
+	efa_free_mapped(dev, cq->cpu_addr, cq->dma_addr, cq->size, DMA_FROM_DEVICE);
+err_free_sub_cq_arr:
+	kfree(cq->sub_cq_arr);
+err_out:
+	atomic64_inc(&dev->stats.create_cq_err);
+	return err;
+}
+#endif
 int efa_create_cq(struct ib_cq *ibcq, const struct ib_cq_init_attr *attr,
+#ifndef HAVE_CREATE_CQ_BUNDLE
 		  struct ib_udata *udata)
 {
+#else
+		  struct uverbs_attr_bundle *attrs)
+{
+	struct ib_udata *udata = &attrs->driver_udata;
+#endif
 #ifdef HAVE_UDATA_TO_DRV_CONTEXT
 	struct efa_ucontext *ucontext = rdma_udata_to_drv_context(
 		udata, struct efa_ucontext, ibucontext);
@@ -1437,12 +1901,9 @@ int efa_create_cq(struct ib_cq *ibcq, const struct ib_cq_init_attr *attr,
 		goto err_out;
 	}
 
-#ifndef HAVE_NO_KVERBS_DRIVERS
-	if (!udata) {
-		ibdev_dbg(ibdev, "udata is NULL\n");
-		err = -EOPNOTSUPP;
-		goto err_out;
-	}
+#ifdef HAVE_EFA_KVERBS
+	if (!udata)
+		return efa_create_cq_kernel(ibcq, attr);
 #endif
 
 	if (offsetofend(typeof(cmd), num_sub_cqs) > udata->inlen) {
@@ -1502,7 +1963,7 @@ int efa_create_cq(struct ib_cq *ibcq, const struct ib_cq_init_attr *attr,
 	}
 
 	params.uarn = cq->ucontext->uarn;
-	params.cq_depth = entries;
+	params.sub_cq_depth = entries;
 	params.dma_addr = cq->dma_addr;
 	params.entry_size_in_bytes = cmd.cq_entry_size;
 	params.num_sub_cqs = cmd.num_sub_cqs;
@@ -1620,7 +2081,6 @@ err_free_cq:
 }
 #endif
 
-#ifdef HAVE_IB_UMEM_FIND_SINGLE_PG_SIZE
 static int umem_to_page_list(struct efa_dev *dev,
 			     struct ib_umem *umem,
 			     u64 *page_list,
@@ -1634,78 +2094,11 @@ static int umem_to_page_list(struct efa_dev *dev,
 	ibdev_dbg(&dev->ibdev, "hp_cnt[%u], pages_in_hp[%u]\n",
 		  hp_cnt, pages_in_hp);
 
-	rdma_umem_for_each_dma_block(umem, &biter, BIT(hp_shift))
+	efa_rdma_umem_for_each_dma_block(umem, &biter, BIT(hp_shift))
 		page_list[hp_idx++] = rdma_block_iter_dma_address(&biter);
 
 	return 0;
 }
-#elif defined(HAVE_SG_DMA_PAGE_ITER)
-static int umem_to_page_list(struct efa_dev *dev,
-			     struct ib_umem *umem,
-			     u64 *page_list,
-			     u32 hp_cnt,
-			     u8 hp_shift)
-{
-	u32 pages_in_hp = BIT(hp_shift - PAGE_SHIFT);
-	struct sg_dma_page_iter sg_iter;
-	unsigned int page_idx = 0;
-	unsigned int hp_idx = 0;
-
-	ibdev_dbg(&dev->ibdev, "hp_cnt[%u], pages_in_hp[%u]\n",
-		  hp_cnt, pages_in_hp);
-
-	for_each_sg_dma_page(umem->sg_head.sgl, &sg_iter, umem->nmap, 0) {
-		if (page_idx % pages_in_hp == 0) {
-			page_list[hp_idx] = sg_page_iter_dma_address(&sg_iter);
-			hp_idx++;
-		}
-
-		page_idx++;
-	}
-
-	return 0;
-}
-#elif defined(HAVE_UMEM_SCATTERLIST_IF)
-static int umem_to_page_list(struct efa_dev *dev,
-			     struct ib_umem *umem,
-			     u64 *page_list,
-			     u32 hp_cnt,
-			     u8 hp_shift)
-{
-	u32 pages_in_hp = BIT(hp_shift - PAGE_SHIFT);
-	unsigned int page_idx = 0;
-	unsigned int pages_in_sg;
-	unsigned int hp_idx = 0;
-	struct scatterlist *sg;
-	unsigned int entry;
-	unsigned int i;
-
-	ibdev_dbg(&dev->ibdev, "hp_cnt[%u], pages_in_hp[%u]\n",
-		  hp_cnt, pages_in_hp);
-
-	for_each_sg(umem->sg_head.sgl, sg, umem->nmap, entry) {
-		if (sg_dma_len(sg) & ~PAGE_MASK) {
-			ibdev_dbg(&dev->ibdev,
-				  "sg_dma_len[%u] does not divide by PAGE_SIZE[%lu]\n",
-				  sg_dma_len(sg), PAGE_SIZE);
-			return -EINVAL;
-		}
-
-		pages_in_sg = sg_dma_len(sg) >> PAGE_SHIFT;
-		for (i = 0; i < pages_in_sg; i++) {
-			if (page_idx % pages_in_hp == 0) {
-				page_list[hp_idx] = sg_dma_address(sg) +
-						    i * PAGE_SIZE;
-				hp_idx++;
-			}
-
-			page_idx++;
-		}
-	}
-
-	return 0;
-}
-#endif
 
 static struct scatterlist *efa_vmalloc_buf_to_sg(u64 *buf, int page_cnt)
 {
@@ -1744,12 +2137,7 @@ static int pbl_chunk_list_create(struct efa_dev *dev, struct pbl_context *pbl)
 	int sg_dma_cnt = pbl->phys.indirect.sg_dma_cnt;
 	struct efa_com_ctrl_buff_info *ctrl_buf;
 	u64 *cur_chunk_buf, *prev_chunk_buf;
-#ifdef HAVE_IB_UMEM_FIND_SINGLE_PG_SIZE
 	struct ib_block_iter biter;
-#else
-	struct scatterlist *sg;
-	unsigned int entry, payloads_in_sg;
-#endif
 	dma_addr_t dma_addr;
 	int i;
 
@@ -1783,7 +2171,6 @@ static int pbl_chunk_list_create(struct efa_dev *dev, struct pbl_context *pbl)
 	chunk_idx = 0;
 	payload_idx = 0;
 	cur_chunk_buf = chunk_list->chunks[0].buf;
-#ifdef HAVE_IB_UMEM_FIND_SINGLE_PG_SIZE
 	rdma_for_each_block(pages_sgl, &biter, sg_dma_cnt,
 			    EFA_CHUNK_PAYLOAD_SIZE) {
 		cur_chunk_buf[payload_idx++] =
@@ -1795,22 +2182,6 @@ static int pbl_chunk_list_create(struct efa_dev *dev, struct pbl_context *pbl)
 			payload_idx = 0;
 		}
 	}
-#else
-	for_each_sg(pages_sgl, sg, sg_dma_cnt, entry) {
-		payloads_in_sg = sg_dma_len(sg) >> EFA_CHUNK_PAYLOAD_SHIFT;
-		for (i = 0; i < payloads_in_sg; i++) {
-			cur_chunk_buf[payload_idx++] =
-				(sg_dma_address(sg) & ~(EFA_CHUNK_PAYLOAD_SIZE - 1)) +
-				(EFA_CHUNK_PAYLOAD_SIZE * i);
-
-			if (payload_idx == EFA_PTRS_PER_CHUNK) {
-				chunk_idx++;
-				cur_chunk_buf = chunk_list->chunks[chunk_idx].buf;
-				payload_idx = 0;
-			}
-		}
-	}
-#endif
 
 	/* map chunks to dma and fill chunks next ptrs */
 	for (i = chunk_list_size - 1; i >= 0; i--) {
@@ -2089,54 +2460,6 @@ static int efa_create_pbl(struct efa_dev *dev,
 	return 0;
 }
 
-#ifndef HAVE_IB_UMEM_FIND_SINGLE_PG_SIZE
-static unsigned long efa_cont_pages(struct ib_umem *umem,
-				    unsigned long page_size_cap,
-				    u64 addr)
-{
-	unsigned long max_page_shift = fls64(page_size_cap);
-	struct scatterlist *sg;
-	u64 base = ~0, p = 0;
-	unsigned long tmp;
-	unsigned long m;
-	u64 len, pfn;
-	int i = 0;
-	int entry;
-
-	addr = addr >> PAGE_SHIFT;
-	tmp = (unsigned long)addr;
-	m = find_first_bit(&tmp, BITS_PER_LONG);
-	m = min_t(unsigned long, max_page_shift - PAGE_SHIFT, m);
-
-	for_each_sg(umem->sg_head.sgl, sg, umem->nmap, entry) {
-		len = DIV_ROUND_UP(sg_dma_len(sg), PAGE_SIZE);
-		pfn = sg_dma_address(sg) >> PAGE_SHIFT;
-		if (base + p != pfn) {
-			/*
-			 * If either the offset or the new
-			 * base are unaligned update m
-			 */
-			tmp = (unsigned long)(pfn | p);
-			if (!IS_ALIGNED(tmp, 1 << m))
-				m = find_first_bit(&tmp, BITS_PER_LONG);
-
-			base = pfn;
-			p = 0;
-		}
-
-		p += len;
-		i += len;
-	}
-
-	if (i)
-		m = min_t(unsigned long, ilog2(roundup_pow_of_two(i)), m);
-	else
-		m = 0;
-
-	return BIT(PAGE_SHIFT + m);
-}
-#endif
-
 static struct efa_mr *efa_alloc_mr(struct ib_pd *ibpd, int access_flags,
 				   struct ib_udata *udata)
 {
@@ -2144,13 +2467,6 @@ static struct efa_mr *efa_alloc_mr(struct ib_pd *ibpd, int access_flags,
 	int supp_access_flags;
 	struct efa_mr *mr;
 
-#ifndef HAVE_NO_KVERBS_DRIVERS
-	if (!udata) {
-		ibdev_dbg(&dev->ibdev, "udata is NULL\n");
-		return ERR_PTR(-EINVAL);
-	}
-#endif
-
 	if (udata && udata->inlen &&
 	    !ib_is_udata_cleared(udata, 0, sizeof(udata->inlen))) {
 		ibdev_dbg(&dev->ibdev,
@@ -2203,19 +2519,14 @@ static int efa_register_mr(struct ib_pd *ibpd, struct efa_mr *mr, u64 start,
 	}
 #endif
 
-#ifdef HAVE_IB_UMEM_FIND_SINGLE_PG_SIZE
-	pg_sz = ib_umem_find_best_pgsz(mr->umem,
-				       dev->dev_attr.page_size_cap,
-				       virt_addr);
+	pg_sz = efa_umem_find_best_pgsz(mr->umem,
+					dev->dev_attr.page_size_cap,
+					virt_addr);
 	if (!pg_sz) {
 		ibdev_dbg(&dev->ibdev, "Failed to find a suitable page size in page_size_cap %#llx\n",
 			  dev->dev_attr.page_size_cap);
 		return -EOPNOTSUPP;
 	}
-#else
-	pg_sz = efa_cont_pages(mr->umem, dev->dev_attr.page_size_cap,
-			       virt_addr);
-#endif /* defined(HAVE_IB_UMEM_FIND_SINGLE_PG_SIZE) */
 
 #ifdef HAVE_EFA_P2P
 skip_umem_pg_sz:
@@ -2264,8 +2575,9 @@ skip_umem_pg_sz:
 
 	mr->ibmr.lkey = result.l_key;
 	mr->ibmr.rkey = result.r_key;
-#ifdef HAVE_IB_MR_LENGTH
 	mr->ibmr.length = length;
+#ifndef HAVE_IB_MR_TYPE
+	mr->type = IB_MR_TYPE_USER;
 #endif
 	mr->ic_info.recv_ic_id = result.ic_info.recv_ic_id;
 	mr->ic_info.rdma_read_ic_id = result.ic_info.rdma_read_ic_id;
@@ -2288,14 +2600,22 @@ skip_umem_pg_sz:
 struct ib_mr *efa_reg_user_mr_dmabuf(struct ib_pd *ibpd, u64 start,
 				     u64 length, u64 virt_addr,
 				     int fd, int access_flags,
+#ifndef HAVE_REG_USER_MR_DMABUF_BUNDLE
 				     struct ib_udata *udata)
+#else
+				     struct uverbs_attr_bundle *attrs)
+#endif
 {
 	struct efa_dev *dev = to_edev(ibpd->device);
 	struct ib_umem_dmabuf *umem_dmabuf;
 	struct efa_mr *mr;
 	int err;
 
+#ifndef HAVE_REG_USER_MR_DMABUF_BUNDLE
 	mr = efa_alloc_mr(ibpd, access_flags, udata);
+#else
+	mr = efa_alloc_mr(ibpd, access_flags, &attrs->driver_udata);
+#endif
 	if (IS_ERR(mr)) {
 		err = PTR_ERR(mr);
 		goto err_out;
@@ -2395,6 +2715,141 @@ err_out:
 	return ERR_PTR(err);
 }
 
+#ifdef HAVE_EFA_KVERBS
+struct ib_mr *efa_get_dma_mr(struct ib_pd *ibpd, int access_flags)
+{
+	struct efa_dev *dev = to_edev(ibpd->device);
+	u64 length = BIT(dev->edev.dma_addr_bits);
+	struct efa_com_reg_mr_params params = {};
+	struct efa_com_reg_mr_result result = {};
+	int supp_access_flags;
+	struct efa_mr *mr;
+	int err;
+
+	supp_access_flags = IB_ACCESS_LOCAL_WRITE;
+
+#ifdef HAVE_IB_ACCESS_OPTIONAL
+	access_flags &= ~IB_ACCESS_OPTIONAL;
+#endif
+
+	if (access_flags & ~supp_access_flags) {
+		ibdev_dbg(&dev->ibdev, "Unsupported access flags[%#x], supported[%#x]\n",
+			  access_flags, supp_access_flags);
+		err = -EOPNOTSUPP;
+		goto err_out;
+	}
+
+	mr = kzalloc(sizeof(*mr), GFP_KERNEL);
+	if (!mr) {
+		err = -ENOMEM;
+		goto err_out;
+	}
+
+	params.pd = to_epd(ibpd)->pdn;
+	params.iova = 0;
+	params.mr_length_in_bytes = length;
+	params.permissions = access_flags;
+	params.page_num = 0;
+
+	err = efa_com_register_mr(&dev->edev, &params, &result);
+	if (err) {
+		ibdev_dbg(&dev->ibdev, "Failed to register dma mr [%d]\n", err);
+		goto err_free;
+	}
+
+	mr->ibmr.lkey = result.l_key;
+	mr->ibmr.rkey = result.r_key;
+	mr->ibmr.length = length;
+#ifndef HAVE_IB_MR_TYPE
+	mr->type = IB_MR_TYPE_DMA;
+#endif
+	ibdev_dbg(&dev->ibdev, "Registered dma mr[%d]\n", mr->ibmr.lkey);
+
+	return &mr->ibmr;
+
+err_free:
+	kfree(mr);
+err_out:
+	atomic64_inc(&dev->stats.get_dma_mr_err);
+	return ERR_PTR(err);
+}
+
+#ifdef HAVE_ALLOC_MR_UDATA
+struct ib_mr *efa_alloc_fast_mr(struct ib_pd *ibpd, enum ib_mr_type mr_type,
+				u32 max_num_sg, struct ib_udata *udata)
+#else
+struct ib_mr *efa_alloc_fast_mr(struct ib_pd *ibpd, enum ib_mr_type mr_type,
+				u32 max_num_sg)
+#endif
+{
+	struct efa_dev *dev = to_edev(ibpd->device);
+	struct efa_com_alloc_mr_params params = {};
+	struct efa_com_alloc_mr_result result = {};
+	dma_addr_t pages_list_dma_addr;
+	u32 buf_size_in_bytes;
+	struct efa_mr *mr;
+	int err;
+
+	if (mr_type != IB_MR_TYPE_MEM_REG) {
+		ibdev_dbg(&dev->ibdev, "Invalid mr type %d\n", mr_type);
+		err = -EINVAL;
+		goto err_out;
+	}
+
+	mr = kzalloc(sizeof(*mr), GFP_KERNEL);
+	if (!mr) {
+		err = -ENOMEM;
+		goto err_out;
+	}
+
+#ifndef HAVE_IB_MR_TYPE
+	mr->type = mr_type;
+#endif
+
+	buf_size_in_bytes = max_num_sg * sizeof(*mr->pages_list);
+	mr->pages_list = kzalloc(buf_size_in_bytes, GFP_KERNEL);
+	if (!mr->pages_list) {
+		err = -ENOMEM;
+		goto err_free_mr;
+	}
+
+	pages_list_dma_addr = dma_map_single(&dev->pdev->dev, mr->pages_list, buf_size_in_bytes,
+					     DMA_TO_DEVICE);
+	if (dma_mapping_error(&dev->pdev->dev, pages_list_dma_addr)) {
+		ibdev_dbg(&dev->ibdev, "Unable to map host pages array to DMA address\n");
+		err = -ENOMEM;
+		goto err_free_pages_list;
+	}
+
+	mr->pages_list_length = max_num_sg;
+	mr->pages_list_dma_addr = pages_list_dma_addr;
+
+	params.max_pages = max_num_sg;
+	params.pd = to_epd(ibpd)->pdn;
+
+	err = efa_com_alloc_mr(&dev->edev, &params, &result);
+	if (err)
+		goto err_unmap_pages;
+
+	mr->ibmr.lkey = result.l_key;
+	mr->ibmr.rkey = result.r_key;
+
+	ibdev_dbg(&dev->ibdev, "Allocated mr[%d]\n", mr->ibmr.lkey);
+
+	return &mr->ibmr;
+
+err_unmap_pages:
+	dma_unmap_single(&dev->pdev->dev, pages_list_dma_addr, buf_size_in_bytes, DMA_TO_DEVICE);
+err_free_pages_list:
+	kfree(mr->pages_list);
+err_free_mr:
+	kfree(mr);
+err_out:
+	atomic64_inc(&dev->stats.alloc_mr_err);
+	return ERR_PTR(err);
+}
+#endif
+
 #ifdef HAVE_IB_DEVICE_DRIVER_DEF
 static int UVERBS_HANDLER(EFA_IB_METHOD_MR_QUERY)(struct uverbs_attr_bundle *attrs)
 {
@@ -2430,6 +2885,21 @@ static int UVERBS_HANDLER(EFA_IB_METHOD_MR_QUERY)(struct uverbs_attr_bundle *att
 }
 #endif
 
+#if defined(HAVE_MR_DMABUF) && !defined(HAVE_IB_UMEM_DMABUF_PINNED)
+static void efa_unpin_dmabuf(struct efa_mr *mr)
+{
+	struct ib_umem_dmabuf *umem_dmabuf;
+
+	if (!mr->umem->is_dmabuf)
+		return;
+
+	umem_dmabuf = to_ib_umem_dmabuf(mr->umem);
+	dma_resv_lock(umem_dmabuf->attach->dmabuf->resv, NULL);
+	dma_buf_unpin(umem_dmabuf->attach);
+	dma_resv_unlock(umem_dmabuf->attach->dmabuf->resv);
+}
+#endif
+
 #ifdef HAVE_DEREG_MR_UDATA
 int efa_dereg_mr(struct ib_mr *ibmr, struct ib_udata *udata)
 #else
@@ -2441,10 +2911,25 @@ int efa_dereg_mr(struct ib_mr *ibmr)
 	struct efa_mr *mr = to_emr(ibmr);
 	int err;
 
+#ifdef HAVE_EFA_KVERBS
+#ifdef HAVE_IB_MR_TYPE
+	enum ib_mr_type mr_type;
+
+	mr_type = ibmr->type;
+#else
+	u8 mr_type;
+
+	mr_type = mr->type;
+#endif
+#endif
 	ibdev_dbg(&dev->ibdev, "Deregister mr[%d]\n", ibmr->lkey);
 
 #ifdef HAVE_EFA_P2P
+#ifdef HAVE_EFA_KVERBS
+	if (mr_type == IB_MR_TYPE_USER && mr->p2pmem) {
+#else
 	if (mr->p2pmem) {
+#endif
 		err = efa_p2p_put(mr->p2p_ticket, false);
 		if (err)
 			return err;
@@ -2458,18 +2943,29 @@ int efa_dereg_mr(struct ib_mr *ibmr)
 	if (err)
 		return err;
 
+#ifdef HAVE_EFA_KVERBS
+	switch (mr_type) {
+	case IB_MR_TYPE_USER:
 #if defined(HAVE_MR_DMABUF) && !defined(HAVE_IB_UMEM_DMABUF_PINNED)
-	if (mr->umem->is_dmabuf) {
-		struct ib_umem_dmabuf *umem_dmabuf;
-
-		umem_dmabuf = to_ib_umem_dmabuf(mr->umem);
-		dma_resv_lock(umem_dmabuf->attach->dmabuf->resv, NULL);
-		dma_buf_unpin(umem_dmabuf->attach);
-		dma_resv_unlock(umem_dmabuf->attach->dmabuf->resv);
+		efa_unpin_dmabuf(mr);
+#endif
+		ib_umem_release(mr->umem);
+		break;
+	case IB_MR_TYPE_MEM_REG:
+		dma_unmap_single(&dev->pdev->dev, mr->pages_list_dma_addr,
+				 mr->pages_list_length * sizeof(*mr->pages_list),
+				 DMA_TO_DEVICE);
+		kfree(mr->pages_list);
+		break;
+	default:
+		break;
 	}
+#else /* !HAVE_EFA_KVERBS */
+#if defined(HAVE_MR_DMABUF) && !defined(HAVE_IB_UMEM_DMABUF_PINNED)
+	efa_unpin_dmabuf(mr);
 #endif
-
 	ib_umem_release(mr->umem);
+#endif /* HAVE_EFA_KVERBS */
 	kfree(mr);
 
 	return 0;
@@ -2789,15 +3285,11 @@ int efa_create_ah(struct ib_ah *ibah,
 	}
 #endif
 
-#ifndef HAVE_NO_KVERBS_DRIVERS
-	if (!udata) {
-		ibdev_dbg(&dev->ibdev, "udata is NULL\n");
-		err = -EOPNOTSUPP;
-		goto err_out;
-	}
-#endif
-
+#ifdef HAVE_EFA_KVERBS
+	if (udata && udata->inlen &&
+#else
 	if (udata->inlen &&
+#endif
 	    !ib_is_udata_cleared(udata, 0, udata->inlen)) {
 		ibdev_dbg(&dev->ibdev, "Incompatible ABI params\n");
 		err = -EINVAL;
@@ -2816,7 +3308,11 @@ int efa_create_ah(struct ib_ah *ibah,
 
 	resp.efa_address_handle = result.ah;
 
+#ifdef HAVE_EFA_KVERBS
+	if (udata && udata->outlen) {
+#else
 	if (udata->outlen) {
+#endif
 		err = ib_copy_to_udata(udata, &resp,
 				       min(sizeof(resp), udata->outlen));
 		if (err) {
@@ -2965,6 +3461,10 @@ static int efa_fill_device_stats(struct efa_dev *dev,
 		atomic64_read(&s->alloc_ucontext_err);
 	stats->value[EFA_CREATE_AH_ERR] = atomic64_read(&s->create_ah_err);
 	stats->value[EFA_MMAP_ERR] = atomic64_read(&s->mmap_err);
+#ifdef HAVE_EFA_KVERBS
+	stats->value[EFA_ALLOC_MR_ERR] = atomic64_read(&s->alloc_mr_err);
+	stats->value[EFA_GET_DMA_MR_ERR] = atomic64_read(&s->get_dma_mr_err);
+#endif
 
 	return ARRAY_SIZE(efa_device_stats_descs);
 }
@@ -3041,66 +3541,6 @@ int efa_get_hw_stats(struct ib_device *ibdev, struct rdma_hw_stats *stats,
 		return efa_fill_device_stats(to_edev(ibdev), stats);
 }
 
-#ifndef HAVE_NO_KVERBS_DRIVERS
-#ifdef HAVE_POST_CONST_WR
-int efa_post_send(struct ib_qp *ibqp,
-		  const struct ib_send_wr *wr,
-		  const struct ib_send_wr **bad_wr)
-#else
-int efa_post_send(struct ib_qp *ibqp,
-		  struct ib_send_wr *wr,
-		  struct ib_send_wr **bad_wr)
-#endif
-{
-	struct efa_dev *dev = to_edev(ibqp->device);
-
-	ibdev_warn(&dev->ibdev.dev, "Function not supported\n");
-	return -EOPNOTSUPP;
-}
-
-#ifdef HAVE_POST_CONST_WR
-int efa_post_recv(struct ib_qp *ibqp,
-		  const struct ib_recv_wr *wr,
-		  const struct ib_recv_wr **bad_wr)
-#else
-int efa_post_recv(struct ib_qp *ibqp,
-		  struct ib_recv_wr *wr,
-		  struct ib_recv_wr **bad_wr)
-#endif
-{
-	struct efa_dev *dev = to_edev(ibqp->device);
-
-	ibdev_warn(&dev->ibdev.dev, "Function not supported\n");
-	return -EOPNOTSUPP;
-}
-
-int efa_poll_cq(struct ib_cq *ibcq, int num_entries,
-		struct ib_wc *wc)
-{
-	struct efa_dev *dev = to_edev(ibcq->device);
-
-	ibdev_warn(&dev->ibdev.dev, "Function not supported\n");
-	return -EOPNOTSUPP;
-}
-
-int efa_req_notify_cq(struct ib_cq *ibcq,
-		      enum ib_cq_notify_flags flags)
-{
-	struct efa_dev *dev = to_edev(ibcq->device);
-
-	ibdev_warn(&dev->ibdev.dev, "Function not supported\n");
-	return -EOPNOTSUPP;
-}
-
-struct ib_mr *efa_get_dma_mr(struct ib_pd *ibpd, int acc)
-{
-	struct efa_dev *dev = to_edev(ibpd->device);
-
-	ibdev_warn(&dev->ibdev.dev, "Function not supported\n");
-	return ERR_PTR(-EOPNOTSUPP);
-}
-#endif
-
 enum rdma_link_layer efa_port_link_layer(struct ib_device *ibdev,
 					 port_t port_num)
 {
